/************************************************************************
 *
 *                    CalC (Calcium Calculator)
 *                         Victor Matveev
 *
 *                            box.cpp
 *
 *               Geometry definition classes:
 * 
 *               Generalized box class BoxObj
 *               Box as an obstacle: class ObstBoxObj
 *               Set of boxes and obstacles: RegionObj
 *
 ************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include "syntax.h"
#include "grid.h"
#include "box.h"

class GridObj *BoxObjClass::Grid = 0;
class GridObj *RegionObj::Grid = 0;


//***************************************************************
//                I M P L E M E N T A T I O N S
//***************************************************************

int BoxObj::inside(int ix, int iy, int iz) {

  if (ix < 0 || ix >= Grid->xsize) return 0;

  bool b = (Grid->xcoord[ix] < xmax + 0.5 * Grid->xgrid[ix - ( ix > 0 ? 1 : 0)]) && 
           (Grid->xcoord[ix] > xmin - 0.5 * Grid->xgrid[ix]);

  if ( DIMENSIONALITY > 1) {
    if (iy < 0 || iy >= Grid->ysize) return 0;
    b = b && (Grid->ycoord[iy] < ymax + 0.5 * Grid->ygrid[iy - ( iy > 0 ? 1 : 0) ]) && 
             (Grid->ycoord[iy] > ymin - 0.5 * Grid->ygrid[iy]);
  }

  if (DIMENSIONALITY > 2) {
     if (iz < 0 || iz >= Grid->zsize) return 0;
     b = b && (Grid->zcoord[iz] < zmax + 0.5 * Grid->zgrid[iz - ( iz > 0 ? 1 : 0) ]) && 
	          (Grid->zcoord[iz] > zmin - 0.5 * Grid->zgrid[iz]);
  }

  if (b) return _INSIDE_;  else return 0;
}

//***************************************************************

int ObstBoxObj::inside(int ix, int iy, int iz)
  {    
  if (ix < 0 || ix >= Grid->xsize) return _INSIDE_;

     bool b = (Grid->xcoord[ix] < xmax - 0.5 * Grid->xgrid[ix]) && 
              (Grid->xcoord[ix] > xmin + 0.5 * Grid->xgrid[ix - ( ix > 0 ? 1 : 0) ]);
  if (DIMENSIONALITY > 1) {
     if (iy < 0 || iy >= Grid->ysize) return _INSIDE_;
     b = b && (Grid->ycoord[iy] < ymax - 0.5 * Grid->ygrid[iy]) && 
	          (Grid->ycoord[iy] > ymin + 0.5 * Grid->ygrid[iy - ( iy > 0 ? 1 : 0) ]);
  }
  if (DIMENSIONALITY > 2) {
     if (iz < 0 || iz >= Grid->zsize) return _INSIDE_;
     b = b && (Grid->zcoord[iz] < zmax - 0.5 * Grid->zgrid[iz]) && 
      	      (Grid->zcoord[iz] > zmin + 0.5 * Grid->zgrid[iz - ( iz > 0 ? 1 : 0) ]);
  }

  if (b) return 0; else return _INSIDE_;
  }

//***************************************************************

int SphereObj::inside(int ix, int iy, int iz) {
    
  if (ix < 0 || ix >= Grid->xsize || iy < 0 || iy >= Grid->ysize || iz < 0 || iz >= Grid->zsize )
    return 0;
  
  double dr = sqrt(Grid->xgrid[ix] * Grid->xgrid[ix] + Grid->ygrid[iy] * Grid->ygrid[iy] + Grid->zgrid[iz] * Grid->zgrid[iz] ); 

  double dist = sqrt( (Grid->xcoord[ix] - xcenter) * (Grid->xcoord[ix] - xcenter) +
                      (Grid->ycoord[iy] - ycenter) * (Grid->ycoord[iy] - ycenter) +
                      (Grid->zcoord[iz] - zcenter) * (Grid->zcoord[iz] - zcenter) );

  if (dist < radius + 0.2*dr)  return _INSIDE_; 
  else return 0;
}

//***************************************************************

int ObstSphereObj::inside(int ix, int iy, int iz) {
    
  
  if (ix < 0 || ix >= Grid->xsize || iy < 0 || iy >= Grid->ysize || iz < 0 || iz >= Grid->zsize )
    return _INSIDE_;
 
  double dr = sqrt(Grid->xgrid[ix] * Grid->xgrid[ix] + Grid->ygrid[iy] * Grid->ygrid[iy] + Grid->zgrid[iz] * Grid->zgrid[iz] ); 

  double dist = sqrt( (Grid->xcoord[ix] - xcenter) * (Grid->xcoord[ix] - xcenter) +
                      (Grid->ycoord[iy] - ycenter) * (Grid->ycoord[iy] - ycenter) +
                      (Grid->zcoord[iz] - zcenter) * (Grid->zcoord[iz] - zcenter) );

  if (dist > radius - 0.2*dr) return _INSIDE_; else return 0;
}


//***************************************************************

int BoxObjClass::point_type(int ix, int iy, int iz)
   {
   int  surface;

   if ( (surface = inside(ix, iy, iz)) > 0 )
      {
      if (! inside(ix - 1, iy, iz) ) surface |= SURF_XMIN;
      if (! inside(ix + 1, iy, iz) ) surface |= SURF_XMAX;
      if (DIMENSIONALITY == 1) return surface;
      if (! inside(ix, iy - 1, iz) ) surface |= SURF_YMIN;
      if (! inside(ix, iy + 1, iz) ) surface |= SURF_YMAX;
      if (DIMENSIONALITY == 2) return surface;
      if (! inside(ix, iy, iz - 1) ) surface |= SURF_ZMIN;
      if (! inside(ix, iy, iz + 1) ) surface |= SURF_ZMAX;
      }

    return surface;
    }


//***************************************************************
//*   C L A S S   R E G I O N  I M P L E M E N T A T I O N S
//***************************************************************

RegionObj::RegionObj(TokenString &pars)
{

 if (VERBOSE) { fprintf(stderr,"\n#### Building the geometry:"); fflush(stderr); }

   encl_num = pars.token_count("volume");
 sphere_num = pars.token_count("sphere");
   obst_num = pars.token_count("obstacle");
  sobst_num = pars.token_count("sobstacle");

 if (!(encl_num+sphere_num)) globalError(StrCpy("No diffusion volume definitions in script file"));

 initialize( encl_num, sphere_num, obst_num, sobst_num );

 int i;
 if   (encl_num) for (i = 1; i <=   encl_num; i++)  add_enclosure(pars);
 if (sphere_num) for (i = 1; i <= sphere_num; i++)  add_sphere   (pars);
 if   (obst_num) for (i = 1; i <=   obst_num; i++)  add_obstacle (pars);
 if  (sobst_num) for (i = 1; i <=  sobst_num; i++)  add_sobstacle(pars);
}

//***************************************************************

 void RegionObj::initialize(int vNum, int sNum, int oNum, int soNum) {  
   encl_cnt   = sphere_cnt = obst_cnt = sobst_cnt = 0;
   encl_num   = vNum; 
   obst_num   = oNum;
   sphere_num = sNum;
   sobst_num  = soNum;
 
   total = encl_num + obst_num + sphere_num + sobst_num;

    enclosures = new BoxObj       [encl_num];
       spheres = new SphereObj    [sphere_num];
     obstacles = new ObstBoxObj   [obst_num];
    sobstacles = new ObstSphereObj[sobst_num];
 }

//***************************************************************

RegionObj::~RegionObj()
{
  delete [] enclosures;
  delete [] obstacles;
  delete [] spheres;
  delete [] sobstacles;

}

//***************************************************************
// point_type bit fields are:
// bit # is:    ...8     |   7    |  6   |  5   |  4   |  3   |  2   |  1  
// meaning:   box number | inside | zmax | zmin | ymax | ymin | xmax | xmin
//***************************************************************


int RegionObj::point_type(int ix, int iy, int iz, int extObstNum, class ObstBoxObj *OBO)
{
int    ptype, boxid = 0;
int    encl_bnd = SURF_MASK;
int    obst_bnd = 0;  
int    i, j, inside = 0;

    for (i = 0; i < encl_num; i++)
      {  
      ptype = enclosures[i].point_type(ix, iy, iz); 
      if (ptype & _INSIDE_)    
        {                                      //  Point is inside if it is
		inside = _INSIDE_;                     //  inside any of the enclosures.
		encl_bnd &= ptype;                     //  Inside point is on a boundary only
		if (ptype  & SURF_MASK )  boxid = i;   //  if it belongs to the same boundary
		}                                      //  of all the enclosures containing it
      }    

    for (int i = 0; i < sphere_num; i++)
      {  
      ptype = spheres[i].point_type(ix, iy, iz);
      if (ptype & _INSIDE_)    
        {                                      //  Point is inside if it is
		inside = _INSIDE_;                     //  inside any of the enclosures.
	    encl_bnd &= ptype;                     //  Inside point is on a boundary only
		if (ptype  & SURF_MASK )               //  if it belongs to the same boundary
		          	boxid = i + encl_num; 
		}                                      //  of all the enclosures containing it
      }

    if (encl_bnd == SURF_MASK)  encl_bnd = 0;  

    for (j = 0; j < obst_num; j++)
      {  
      ptype = obstacles[j].point_type(ix, iy, iz);
      if (ptype == 0) inside = 0;              // Point is outside if it belongs
        else                                   // to any of the obstacles.
          if (ptype  & SURF_MASK )  
            { 
            obst_bnd |= (ptype & SURF_MASK);    
            boxid = j + encl_num + sphere_num; 
            }                                 
      }                                       
    
	for (j = 0; j < sobst_num; j++)
      {  
      ptype = sobstacles[j].point_type(ix, iy, iz);
      if (ptype == 0) inside = 0;              // Point is outside if it belongs
        else                                   // to any of the obstacles.
          if (ptype  & SURF_MASK )  
            {                                  
            obst_bnd |= (ptype & SURF_MASK);                 
            boxid = j + encl_num + sphere_num + obst_num;
            }                                 
      }                                        
    

    if (extObstNum)
    for (j = 0; j < extObstNum; j++)
      {  
      ptype = OBO[j].point_type(ix, iy, iz);
      if (ptype == 0) inside = 0;              // Point is outside if it belongs
        else                                   // to any of the obstacles.
          if (ptype  & SURF_MASK )  
            {                                  
            obst_bnd |= (ptype & SURF_MASK);                 
            boxid = j + encl_num + sphere_num + obst_num + sobst_num; 
            }                                 
      }                                        

    if ( inside == 0 ) return 0;

    int bnd = encl_bnd | obst_bnd;
 
    if (bnd == 0) return inside;
    /*
    if ( ((bnd & SURF_XMIN) && (bnd & SURF_XMAX)) ||
         ((bnd & SURF_YMIN) && (bnd & SURF_YMAX)) ||
         ((bnd & SURF_ZMIN) && (bnd & SURF_ZMAX)) )
        throw makeMessage("Ambiguity: two surfaces touch. Please enlarge the obstacle dimensions, extending it beyond the domain boundary, to avoid surface ambiguity");
    else */  
		return  _INSIDE_ + (boxid << 7) + bnd;
   
}


//***************************************************************

int RegionObj::point_type(long ind, int extObstNum, class ObstBoxObj *OBO)
{
int iz = ind / Grid->xysize;
ind -= iz * Grid->xysize;
int iy = ind / Grid->xsize;
int ix = ind - iy * Grid->xsize;
return point_type(ix, iy, iz, extObstNum, OBO);
}

//***************************************************************

void RegionObj::add_enclosure(double x0, double x1, double y0, double y1,
                              double z0, double z1)
{
 if (encl_cnt >= encl_num) 
   throw makeMessage("RegionObj::add_enclosure: encl_cnt=%d >= encl_num=%d\n",encl_num, encl_cnt);

 if (encl_cnt == 0) 
   { xmin = x0; xmax = x1; ymin = y0; ymax = y1; zmin = z0; zmax = z1; }
 else
   {
   if (x0 < xmin) xmin = x0;  if (x1 > xmax) xmax = x1;
   if (y0 < ymin) ymin = y0;  if (y1 > ymax) ymax = y1;
   if (z0 < zmin) zmin = z0;  if (z1 > zmax) zmax = z1;
   }

 if (VERBOSE) { fprintf(stderr,"   adding enclosure #%d:", encl_cnt+1); fflush(stderr); }

 enclosures[encl_cnt] = BoxObj(x0, x1, y0, y1, z0, z1);

 if (VERBOSE) {
   fprintf(stderr," [%g,%g]", x0, x1);
   if (DIMENSIONALITY > 1) fprintf(stderr, " x [%g,%g]", y0, y1);
   if (DIMENSIONALITY > 2) fprintf(stderr, " x [%g,%g]", z0, z1);
   fprintf(stderr, "\n");
 }

 if ( x0 > x1 || y0 > y1 || z0 > z1 )  throw StrCpy("Check enclosure volume dimensions");

 encl_cnt++;
}

//***************************************************************

void RegionObj::add_enclosure(TokenString &pars)
{
  double x0, x1, y0, y1, z0, z1;
  long   pos;

  try {
    pos = pars.token_index("volume", encl_cnt+1);
    if (DIMENSIONALITY == 1)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'E');
      add_enclosure(x0, x1, 0.0, 0.0, 0.0, 0.0);
      }
    else  if (DIMENSIONALITY == 2)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'E');
      add_enclosure(x0, x1, y0, y1, 0.0, 0.0);
      }
    else
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'd', &z0, 'd', &z1, 'E');    
      add_enclosure(x0, x1, y0, y1, z0, z1);
      }
  }
  catch(char *message) 
    { pars.errorMessage( pos + 1, message); }
}

//***************************************************************

void RegionObj::add_obstacle(double x0, double x1, double y0, double y1, 
                             double z0, double z1)
{
 if (obst_cnt >= obst_num)  
   throw makeMessage("RegionObj::add_obstacle: obst_cnt=%d >= obst_num=%d\n", obst_cnt, obst_num);

 if (VERBOSE) { fprintf(stderr,"   adding obstacle #%d:", obst_cnt+1); fflush(stderr); }

 obstacles[obst_cnt] = ObstBoxObj(x0, x1, y0, y1, z0, z1);

 if (VERBOSE)  {
   fprintf(stderr," [%g,%g]", x0, x1);
   if (DIMENSIONALITY > 1) fprintf(stderr, " x [%g,%g]", y0, y1);
   if (DIMENSIONALITY > 2) fprintf(stderr, " x [%g,%g]", z0, z1);
   fprintf(stderr, "\n");
 }

 if ( x0 > x1 || y0 > y1 || z0 > z1 )   throw StrCpy("Check diffusion obstacle dimensions"); 

 obst_cnt++;
}

//***************************************************************

void RegionObj::add_obstacle(TokenString &pars)
{
  double x0, x1, y0, y1, z0, z1;
  long   pos;

  try {
    pos = pars.token_index("obstacle", obst_cnt+1);
    if (DIMENSIONALITY == 1)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'E');
      add_obstacle(x0, x1, 0.0, 0.0, 0.0, 0.0);
      }
    else  if (DIMENSIONALITY == 2)
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'E');
      add_obstacle(x0, x1, y0, y1, 0.0, 0.0);
      }
    else  // 3 spatial dimensions
      {
      pars.trail_pars(pos, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'd', &z0, 'd', &z1, 'E');    
      add_obstacle(x0, x1, y0, y1, z0, z1);
      }
    }
  catch(char *message)
    { pars.errorMessage( pos + 1, message); }
}



//***************************************************************

void RegionObj::add_sphere(double x0, double y0, double z0, double r)
{
 if (sphere_cnt >= sphere_num) 
   throw makeMessage("RegionObj::add_sphere: sphere_cnt=%d >= sphere_num=%d\n", sphere_cnt, sphere_num);

 if (sphere_cnt == 0) 
   { xmin = x0-r; xmax = x0+r; ymin = y0-r; ymax = y0+r; zmin = z0-r; zmax = z0+r; }
 else
   {
   if (x0-r < xmin) xmin = x0-r;  if (x0+r > xmax) xmax = x0+r;
   if (y0-r < ymin) ymin = y0-r;  if (y0+r > ymax) ymax = y0+r;
   if (z0-r < zmin) zmin = z0-r;  if (z0+r > zmax) zmax = z0+r;
   }

 if (VERBOSE) { fprintf(stderr,"   adding sphere #%d:", sphere_cnt+1); fflush(stderr); }

 spheres[sphere_cnt] = SphereObj(x0, y0, z0, r);

 if (VERBOSE) fprintf(stderr," [%g,%g,%g] radius=%g\n", x0, y0,z0,r);

 if ( r < 0 ) throw StrCpy("Check sphere volume dimensions");

 encl_cnt++;
}

//***************************************************************

void RegionObj::add_sphere(TokenString &pars)
{
  double x0, y0, z0, r;
  long   pos;

  try {
    pos = pars.token_index("sphere", sphere_cnt+1);
    pars.trail_pars(pos, 'd', &x0, 'd', &y0, 'd', &z0, 'd', &r, 'E');    
    add_sphere(x0, y0, z0, r);
  }
  catch(char *message)  
    { pars.errorMessage( pos + 1, message); }
}

//***************************************************************

void RegionObj::add_sobstacle(double x0, double y0, double z0, double r)
{

 if (sobst_cnt >= sobst_num)  
   throw makeMessage("RegionObj::add_obstacle: obst_cnt=%d >= obst_num=%d\n", sobst_cnt, sobst_num);

 if (VERBOSE) { fprintf(stderr,"   adding obstacle #%d:", sobst_cnt+1); fflush(stderr); }

 sobstacles[sobst_cnt] = ObstSphereObj(x0, y0, z0, r);

 if (VERBOSE) fprintf(stderr," [%g,%g,%g] radius=%g\n", x0, y0, z0, r);

 if ( r < 0 )  throw StrCpy("Check spherical obstacle dimensions");

 encl_cnt++;
}

//***************************************************************

void RegionObj::add_sobstacle(TokenString &pars)
{
  double x0, y0, z0, r;
  long   pos;

  try {
    pos = pars.token_index("sobstacle", sobst_cnt+1);
    pars.trail_pars(pos, 'd', &x0, 'd', &y0, 'd', &z0, 'd', &r, 'E');    
    add_sobstacle(x0, y0, z0, r);
  }
  catch(char *message)  
    { pars.errorMessage( pos + 1, message); }
}
//***************************************************************

void print_type(int x)
{
if (x == 0) fprintf(stderr,"[OUT] ");
else
  {
  if (x & _INSIDE_) fprintf(stderr,"[IN");
  if (!(x & SURF_MASK)) fprintf(stderr,"] ");
    else
      {
      if (x & SURF_XMIN) fprintf(stderr," XMIN");
      if (x & SURF_XMAX) fprintf(stderr," XMAX");
      if (x & SURF_YMIN) fprintf(stderr," YMIN");
      if (x & SURF_YMAX) fprintf(stderr," YMAX");
      if (x & SURF_ZMIN) fprintf(stderr," ZMIN");
      if (x & SURF_ZMAX) fprintf(stderr," ZMAX");
      fprintf(stderr," %d] ",x>>9);
      }
  }
}
#ifdef CALC_DEBUG
#endif
//***************************************************************
//                       T H E   E N D
//***************************************************************

